<?php

/*
 * This file is part of the Doctum utility.
 *
 * (c) Fabien Potencier <fabien@symfony.com>
 *
 * This source file is subject to the MIT license that is bundled
 * with this source code in the file LICENSE.
 */

namespace Doctum\Parser;

use PhpParser\Node as AbstractNode;
use PhpParser\Node\Name\FullyQualified;
use PhpParser\NodeVisitorAbstract;
use PhpParser\Node\Stmt\ClassConst as ClassConstNode;
use PhpParser\Node\Stmt\ClassMethod as ClassMethodNode;
use PhpParser\Node\Stmt\Class_ as ClassNode;
use PhpParser\Node\Stmt\ClassLike as ClassLikeNode;
use PhpParser\Node\Stmt\Interface_ as InterfaceNode;
use PhpParser\Node\Stmt\Namespace_ as NamespaceNode;
use PhpParser\Node\Stmt\Property as PropertyNode;
use PhpParser\Node\Stmt\TraitUse as TraitUseNode;
use PhpParser\Node\Stmt\Trait_ as TraitNode;
use PhpParser\Node\Stmt\Use_ as UseNode;
use PhpParser\Node\NullableType;
use Doctum\Project;
use Doctum\Reflection\ClassReflection;
use Doctum\Reflection\ConstantReflection;
use Doctum\Reflection\MethodReflection;
use Doctum\Reflection\ParameterReflection;
use Doctum\Reflection\PropertyReflection;

class NodeVisitor extends NodeVisitorAbstract
{
    protected $context;

    public function __construct(ParserContext $context)
    {
        $this->context = $context;
    }

    public function enterNode(AbstractNode $node)
    {
        if ($node instanceof NamespaceNode) {
            $this->context->enterNamespace($node->name->__toString());
        } elseif ($node instanceof UseNode) {
            $this->addAliases($node);
        } elseif ($node instanceof InterfaceNode) {
            $this->addInterface($node);
        } elseif ($node instanceof ClassNode) {
            $this->addClass($node);
        } elseif ($node instanceof TraitNode) {
            $this->addTrait($node);
        } elseif ($this->context->getClass() && $node instanceof TraitUseNode) {
            $this->addTraitUse($node);
        } elseif ($this->context->getClass() && $node instanceof PropertyNode) {
            $this->addProperty($node);
        } elseif ($this->context->getClass() && $node instanceof ClassMethodNode) {
            $this->addMethod($node);
        } elseif ($this->context->getClass() && $node instanceof ClassConstNode) {
            $this->addConstant($node);
        }
        return null;
    }

    public function leaveNode(AbstractNode $node)
    {
        if ($node instanceof NamespaceNode) {
            $this->context->leaveNamespace();
        } elseif ($node instanceof ClassNode || $node instanceof InterfaceNode || $node instanceof TraitNode) {
            $this->context->leaveClass();
        }
        return null;
    }

    protected function addAliases(UseNode $node)
    {
        foreach ($node->uses as $use) {
            $this->context->addAlias($use->alias !== null ? $use->alias->__toString() : null, $use->name->__toString());
        }
    }

    protected function addInterface(InterfaceNode $node)
    {
        $class = $this->addClassOrInterface($node);

        $class->setInterface(true);
        foreach ($node->extends as $interface) {
            $class->addInterface((string) $interface);
        }
    }

    protected function addClass(ClassNode $node)
    {
        // Skip anonymous classes
        if ($node->isAnonymous()) {
            return;
        }

        $class = $this->addClassOrInterface($node);

        foreach ($node->implements as $interface) {
            $class->addInterface((string) $interface);
        }

        if ($node->extends) {
            $class->setParent((string) $node->extends);
        }
    }

    protected function addTrait(TraitNode $node)
    {
        $class = $this->addClassOrInterface($node);

        $class->setTrait(true);
    }

    protected function addClassOrInterface(ClassLikeNode $node)
    {
        $class = new ClassReflection((string) $node->namespacedName, $node->getLine());
        if ($node instanceof ClassNode) {
            $class->setModifiers($node->flags);
        }
        $class->setNamespace($this->context->getNamespace());
        $class->setAliases($this->context->getAliases());
        $class->setHash($this->context->getHash());
        $class->setFile($this->context->getFile());

        $comment = $this->context->getDocBlockParser()->parse($node->getDocComment(), $this->context, $class);
        $class->setDocComment($node->getDocComment());
        $class->setShortDesc($comment->getShortDesc());
        $class->setLongDesc($comment->getLongDesc());
        $class->setSee($this->resolveSee($comment->getTag('see')));
        if ($errors = $comment->getErrors()) {
            $class->setErrors($errors);
        } else {
            $class->setTags($comment->getOtherTags());
        }

        if ($this->context->getFilter()->acceptClass($class)) {
            if ($errors) {
                $this->context->addErrors((string) $class, $node->getLine(), $errors);
            }
            $this->context->enterClass($class);
        }

        return $class;
    }

    protected function addMethod(ClassMethodNode $node)
    {
        $method = new MethodReflection($node->name->__toString(), $node->getLine());
        $method->setModifiers($node->flags);
        $method->setByRef((string) $node->byRef);

        foreach ($node->params as $param) {
            $parameter = new ParameterReflection($param->var->name, $param->getLine());
            $parameter->setModifiers($param->type);
            $parameter->setByRef($param->byRef);
            if ($param->default) {
                $parameter->setDefault($this->context->getPrettyPrinter()->prettyPrintExpr($param->default));
            }

            $parameter->setVariadic($param->variadic);

            $type = $param->type;
            $typeStr = null;

            if ($param->type !== null && ! $param->type instanceof NullableType) {
                $typeStr = (string) $param->type;
            } elseif ($param->type instanceof NullableType) {
                $type = $param->type->type;
                $typeStr = (string) $param->type->type;
            }

            if ($type instanceof FullyQualified && 0 !== strpos($typeStr, '\\')) {
                $typeStr = '\\' . $typeStr;
            }

            if (null !== $typeStr) {
                $typeArr = [[$typeStr, false]];

                if ($param->type instanceof NullableType) {
                    $typeArr[] = ['null', false];
                }

                $parameter->setHint($this->resolveHint($typeArr));
            }

            $method->addParameter($parameter);
        }

        $comment = $this->context->getDocBlockParser()->parse($node->getDocComment(), $this->context, $method);
        $method->setDocComment($node->getDocComment());
        $method->setShortDesc($comment->getShortDesc());
        $method->setLongDesc($comment->getLongDesc());
        $method->setSee($this->resolveSee($comment->getTag('see')));
        if (!$errors = $comment->getErrors()) {
            $errors = $this->updateMethodParametersFromTags($method, $comment->getTag('param'));

            if ($tag = $comment->getTag('return')) {
                $method->setHint(is_array($tag[0][0]) ? $this->resolveHint($tag[0][0]) : $tag[0][0]);
                $method->setHintDesc($tag[0][1]);
            }

            $method->setExceptions($comment->getTag('throws'));
            $method->setTags($comment->getOtherTags());
        }

        $method->setErrors($errors);

        $returnType = $node->getReturnType();
        $returnTypeStr = null;

        if ($returnType !== null && ! $returnType instanceof NullableType) {
            $returnTypeStr = (string) $returnType;
        } elseif ($returnType instanceof NullableType) {
            $returnTypeStr = (string) $returnType->type;
        }

        if ($returnType instanceof FullyQualified && 0 !== strpos($returnTypeStr, '\\')) {
            $returnTypeStr = '\\' . $returnTypeStr;
        }

        if (null !== $returnTypeStr) {
            $returnTypeArr = [[$returnTypeStr, false]];

            if ($returnType instanceof NullableType) {
                $returnTypeArr[] = ['null', false];
            }

            $method->setHint($this->resolveHint($returnTypeArr));
        }

        if ($this->context->getFilter()->acceptMethod($method)) {
            $this->context->getClass()->addMethod($method);

            if ($errors) {
                $this->context->addErrors((string) $method, $node->getLine(), $errors);
            }
        }
    }

    protected function addProperty(PropertyNode $node)
    {
        foreach ($node->props as $prop) {
            $property = new PropertyReflection($prop->name, $prop->getLine());
            $property->setModifiers($node->flags);

            $property->setDefault($prop->default);

            $comment = $this->context->getDocBlockParser()->parse($node->getDocComment(), $this->context, $property);
            $property->setDocComment($node->getDocComment());
            $property->setShortDesc($comment->getShortDesc());
            $property->setLongDesc($comment->getLongDesc());
            $property->setSee($this->resolveSee($comment->getTag('see')));
            if ($errors = $comment->getErrors()) {
                $property->setErrors($errors);
            } else {
                if ($tag = $comment->getTag('var')) {
                    $property->setHint($this->resolveHint($tag[0][0]));
                    $property->setHintDesc($tag[0][1]);
                }

                $property->setTags($comment->getOtherTags());
            }

            if ($this->context->getFilter()->acceptProperty($property)) {
                $this->context->getClass()->addProperty($property);

                if ($errors) {
                    $this->context->addErrors((string) $property, $prop->getLine(), $errors);
                }
            }
        }
    }

    protected function addTraitUse(TraitUseNode $node)
    {
        foreach ($node->traits as $trait) {
            $this->context->getClass()->addTrait((string) $trait);
        }
    }

    protected function addConstant(ClassConstNode $node)
    {
        foreach ($node->consts as $const) {
            $constant = new ConstantReflection($const->name, $const->getLine());
            $comment = $this->context->getDocBlockParser()->parse($node->getDocComment(), $this->context, $constant);
            $constant->setDocComment($node->getDocComment());
            $constant->setShortDesc($comment->getShortDesc());
            $constant->setLongDesc($comment->getLongDesc());

            $this->context->getClass()->addConstant($constant);
        }
    }

    protected function updateMethodParametersFromTags(MethodReflection $method, array $tags): array
    {
        // bypass if there is no @param tags defined (@param tags are optional)
        if (!count($tags)) {
            return [];
        }

        if (count($method->getParameters()) != count($tags)) {
            return [sprintf('"%d" @param tags are expected but only "%d" found', count($method->getParameters()), count($tags))];
        }

        $errors = [];
        foreach (array_keys($method->getParameters()) as $i => $name) {
            if ($tags[$i][1] && $tags[$i][1] != $name) {
                $errors[] = sprintf('The "%s" @param tag variable name is wrong (should be "%s")', $tags[$i][1], $name);
            }
        }

        if ($errors) {
            return $errors;
        }

        foreach ($tags as $i => $tag) {
            $parameter = $method->getParameter($tag[1] ? $tag[1] : $i);
            $parameter->setShortDesc($tag[2]);
            if (!$parameter->hasHint()) {
                $parameter->setHint($this->resolveHint($tag[0]));
            }
        }

        return [];
    }

    protected function resolveHint(array $hints): array
    {
        foreach ($hints as $i => $hint) {
            $hints[$i] = [$this->resolveAlias($hint[0]), $hint[1]];
        }

        return $hints;
    }

    protected function resolveAlias($alias)
    {
        // not a class
        if (Project::isPhpTypeHint($alias)) {
            return $alias;
        }

        // FQCN
        if ('\\' == substr($alias, 0, 1)) {
            return $alias;
        }

        $class = $this->context->getClass();

        // A class MIGHT or MIGHT NOT be present in context.
        // It is not present in cases, where eg. `@see` tag refers to non existing class/method.
        // We may want to run class related checks only, if class is actually present.
        if ($class) {
            // special aliases
            if ('self' === $alias || 'static' === $alias || '\$this' === $alias) {
                return $class->getName();
            }

            // an alias defined by a use statement
            $aliases = $class->getAliases();

            if (isset($aliases[$alias])) {
                return $aliases[$alias];
            }

            // a class in the current class namespace
            return $class->getNamespace() . '\\' . $alias;
        }

        return $alias;
    }

    protected function resolveSee(array $see)
    {
        $return = [];
        $matches = [];

        foreach ($see as $seeEntry) {
            $reference = $seeEntry[1];
            $description = $seeEntry[2];
            if ((bool) preg_match('/^[\w]+:\/\/.+$/', $reference)) { //URL
                $return[] = [
                    $reference,
                    $description,
                    false,
                    false,
                    $reference,
                ];
            } elseif ((bool) preg_match('/(.+)\:\:(.+)\(.*\)/', $reference, $matches)) { //Method
                $return[] = [
                    $reference,
                    $description,
                    $this->resolveAlias($matches[1]),
                    $matches[2],
                    false,
                ];
            } else { // We assume, that this is a class reference.
                $return[] = [
                    $reference,
                    $description,
                    $this->resolveAlias($reference),
                    false,
                    false,
                ];
            }
        }

        return $return;
    }
}
